#include <petscdmda.h>
#include <petscksp.h>
#include <petscpcmg.h>
#include <mpi.h>
#include "ComputeFunctions.h"
#include "GlobalVariables.h"

// -- 0.0 --  "ComputerFunctions" Global Functions
extern PetscErrorCode BuildNodes();
extern PetscErrorCode CommUpdate();
extern PetscErrorCode SetUpMPIDerivedDataType();


#undef __FUNCT__
#define __FUNCT__ "Initialization"

PetscErrorCode Initialization(DM da, int time) {
  DM             cda;
  PetscErrorCode ierr;
  Vec gc;

  PetscFunctionBegin;
// -- 1.0 -- Read Info. of DMDA
  ierr = DMDAGetInfo(da, 0, &mx, &my, &mz, &mm , &nn, &pp,0,0,0,0,0,0);CHKERRQ(ierr);
  MPI_Comm_rank(PETSC_COMM_WORLD, &rank);
  ierr = DMDAGetCorners(da,&xs,&ys,&zs,&xw,&yw,&zw);CHKERRQ(ierr);
// -- 1.0 -- Finished

// -- 2.0 -- Generating the grid
  ierr = DMDAGetCoordinateDA(da, &cda);
  ierr = DMDAGetGhostedCoordinates(da, &gc);
  ierr = DMDAVecGetArray(cda, gc, &coors);

  if(time == 0) {
    ierr = BuildNodes();
    ierr = SetUpMPIDerivedDataType();

  } else { printf("Reinitialization ERROR!!\n"); }

// -- Do Not forget to release coors (from 1.1)--
  ierr = DMDAVecRestoreArray(cda, gc, &coors);
  
  PetscFunctionReturn(0);

}

#undef __FUNCT__
#define __FUNCT__ "SetUpMPIDerivedDataType"
PetscErrorCode SetUpMPIDerivedDataType() {
  int blockcounts[1];
  int blocklengths[xc+xg*2+xb*2], displacements[xc+xg*2+xb*2];
  MPI_Aint offsets[1];
  MPI_Datatype  oldtypes[1];


  PetscFunctionBegin;

  rowsize = xc+xg*2+xb*2;
  columnsize = yc+yg*2+yb*2;
  spansize = zc+zg*2+zb*2;

  for(i = 0; i < rowsize; i++) {
    blocklengths[i] = spansize;
    displacements[i] = columnsize*spansize*i;
  }
  MPI_Type_contiguous(columnsize*spansize, MPI_DOUBLE, &rowtype);
  MPI_Type_vector(rowsize*columnsize, 1, spansize, MPI_DOUBLE, &spantype);
  MPI_Type_indexed(rowsize, blocklengths, displacements, MPI_DOUBLE, &columntype);

  MPI_Type_commit(&columntype);
  MPI_Type_commit(&rowtype);
  MPI_Type_commit(&spantype);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "BuildNodes"

PetscErrorCode BuildNodes() {
  int counti, countj, countk;  

  PetscFunctionBegin;

  // -- 1.0 -- Determine the size of local computational grid
  xc = xw;
  yc = yw;
  zc = zw;
    // if it's Neumann BC, you do not need to reduce the computational grid size, i.e.
    // for a 10*10 grid with IB pressure BC's: left, bot, top = Neumann, and right = Dirichlet
    // left: (i=0) == (i=-1); right: (i=imax-1) = contant;

  NbrRank.l = rank-1;
  NbrRank.r = rank+1;
  NbrRank.b = rank-mm;
  NbrRank.t = rank+mm;
  NbrRank.a = rank-mm*nn;
  NbrRank.f = rank+mm*nn; 
  if(xs == 0) {
    localBC.lBC = 1;
  }
  else localBC.lBC = 0;
  if(xs+xw == mx) {
    localBC.rBC = 1;
    xc--;
  }
  else localBC.rBC = 0;
  if(ys == 0) {
    localBC.bBC = 1;
  }
  else localBC.bBC = 0;
  if(ys+yw == my) {
    localBC.tBC = 1;
  }
  else localBC.tBC = 0;
  if(zs == 0) {
    localBC.aBC = 1;
  }
  else localBC.aBC = 0;
  if(zs+zw == mz) {
    localBC.fBC = 1;
  }
  else localBC.fBC = 0;

  printf("rank = %d, NbrRanks are: L: %d, R: %d, B:%d, T:%d, A:%d, F:%d!!\n", rank, NbrRank.l, NbrRank.r, NbrRank.b, NbrRank.t, NbrRank.a, NbrRank.f);
  printf("rank = %d: xs = %d, ys = %d, zs = %d, mx = %d, my = %d, mz = %d!\n", rank, xs, ys, zs, mx, my, mz);

  printf("rank = %d:  left BC = %d,  rightBC = %d, bottom BC = %d, top BC = %d, aft BC = %d, front BC = %d \n", rank, localBC.lBC, localBC.rBC, localBC.bBC, localBC.tBC, localBC.aBC, localBC.fBC);
  printf("rank = %d:  xc = %d, yc = %d, zc = %d, xw = %d, yw = %d, zw = %d\n", rank, xc, yc, zc, xw, yw, zw);

  // -- 1.0 -- Finished
  // in x-direction
  // 0 ~ xg-1 --> left ghosted nodes;                   xg+2*xb+xc ~ xg*2+xb*2+xc --> right ghosted nodes
  _Index.lgEnd = xg-1; // left ghost ending pt
  _Index.rgStr = xg+2*xb+xc; // right ghost starting pt
  // xg ~ xg+xb-1 --> left boundary nodes;              xg+xb+xc ~ xg+2*xb+xc-1 --> right boundary nodes
  _Index.lbEnd = xg+xb-1;
  _Index.rbStr = xg+xb+xc;
  // xg+xb ~ xg+xb+xc-1 --> computational nodes; 
  _Index.lcStr = xg+xb;
  _Index.rcEnd = xg+xb+xc-1;

  // 0 ~ yg-1 --> bottom ghosted nodes;                   yg+2*yb+yc ~ yg*2+yb*2+yc --> top ghosted nodes
  _Index.bgEnd = yg-1;
  _Index.tgStr = yg+2*yb+yc;
  // yg ~ yg+yb-1 --> bottom boundary nodes;              yg+yb+yc ~ yg+2*yb+yc-1 --> top boundary nodes
  _Index.bbEnd = yg+yb-1;
  _Index.tbStr = yg+yb+yc;
  // yg+yb ~ yg+yb+yc-1 --> computational nodes; 
  _Index.bcStr = yg+yb;
  _Index.tcEnd = yg+yb+yc-1;

  // 0 ~ zg-1 --> bottom ghosted nodes;                   zg+2*zb+zc ~ zg*2+zb*2+zc --> top ghosted nodes
  _Index.agEnd = zg-1;
  _Index.fgStr = zg+2*zb+zc;
  // zg ~ zg+zb-1 --> bottom boundary nodes;              zg+zb+zc ~ zg+2*zb+zc-1 --> top boundary nodes
  _Index.abEnd = zg+zb-1;
  _Index.fbStr = zg+zb+zc;
  // zg+zb ~ zg+zb+zc-1 --> computational nodes; 
  _Index.acStr = zg+zb;
  _Index.fcEnd = zg+zb+zc-1;


  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CommUpdate"

PetscErrorCode CommUpdate() {
  MPI_Status status;
  MPI_Request request;
  CommunicationTag SendTag, RecvTag;

  PetscFunctionBegin;


  if(NumRank != 1) {
    printf("rowsize = %d, columnsize = %d, spansize = %d \n", rowsize, columnsize, spansize);
    // 2nd layer always the outer layer;
    SendTag.L1st = 10;
    SendTag.L2nd = 20;
    SendTag.R1st = 30;
    SendTag.R2nd = 40;
    SendTag.T1st = 50;
    SendTag.T2nd = 60;
    SendTag.B1st = 70;
    SendTag.B2nd = 80;
    SendTag.F1st = 90;
    SendTag.F2nd = 100;
    SendTag.A1st = 110;
    SendTag.A2nd = 120;

    RecvTag.L1st = 40;
    RecvTag.L2nd = 30;
    RecvTag.R1st = 20;
    RecvTag.R2nd = 10;
    RecvTag.T1st = 80;
    RecvTag.T2nd = 70;
    RecvTag.B1st = 60;
    RecvTag.B2nd = 50;
    RecvTag.F1st = 120;
    RecvTag.F2nd = 110;
    RecvTag.A1st = 100;
    RecvTag.A2nd = 90;

    // rowsize --> x direction;
    // columntype --> y direction;
    // spantype --> z direction;
    MPI_Barrier(PETSC_COMM_WORLD);

    double _TestV[rowsize][columnsize][spansize];
    for(i = 0; i < xc+xg*2+xb*2; i++) {
      for(j = 0; j < yc+yg*2+yb*2; j++) {
	for(k = 0; k < zc+zg*2+zb*2; k++) {
	  _TestV[i][j][k] = 1.0;
	}
      }
    }
/*
    if(!localBC.tBC) {
      MPI_Send(&_TestV[0][_Index.tcEnd-1][0], 1, columntype, NbrRank.t, SendTag.T1st, PETSC_COMM_WORLD);
      MPI_Send(&_TestV[0][_Index.tcEnd][0], 1, columntype, NbrRank.t, SendTag.T2nd, PETSC_COMM_WORLD);
    }

    if(!localBC.bBC) {
      MPI_Send(&_TestV[0][_Index.bcStr+1][0], 1, columntype, NbrRank.b, SendTag.B1st, PETSC_COMM_WORLD);
      MPI_Send(&_TestV[0][_Index.bcStr][0], 1, columntype, NbrRank.b, SendTag.B2nd, PETSC_COMM_WORLD);
    }

    MPI_Barrier(PETSC_COMM_WORLD);
    printf("Rank = %d finished sending!!!\n", rank);

    if(!localBC.tBC) {
      MPI_Recv(&_TestV[0][_Index.tbStr+1][0], 1, columntype, NbrRank.t, RecvTag.T2nd, PETSC_COMM_WORLD, &status);
      MPI_Recv(&_TestV[0][_Index.tbStr][0], 1, columntype,  NbrRank.t, RecvTag.T1st, PETSC_COMM_WORLD, &status);
    }

    if(!localBC.bBC) {
      MPI_Recv(&_TestV[0][_Index.bbEnd-1][0], 1, columntype, NbrRank.b, RecvTag.B2nd, PETSC_COMM_WORLD, &status);
      MPI_Recv(&_TestV[0][_Index.bbEnd][0], 1, columntype, NbrRank.b, RecvTag.B1st, PETSC_COMM_WORLD, &status);
    }
*/


    if(!localBC.tBC) {
      MPI_Isend(&_TestV[0][_Index.tcEnd-1][0], 1, columntype, NbrRank.t, SendTag.T1st, PETSC_COMM_WORLD, &request);
      MPI_Isend(&_TestV[0][_Index.tcEnd][0], 1, columntype, NbrRank.t, SendTag.T2nd, PETSC_COMM_WORLD, &request);
    }

    if(!localBC.bBC) {
      MPI_Isend(&_TestV[0][_Index.bcStr+1][0], 1, columntype, NbrRank.b, SendTag.B1st, PETSC_COMM_WORLD, &request);
      MPI_Isend(&_TestV[0][_Index.bcStr][0], 1, columntype, NbrRank.b, SendTag.B2nd, PETSC_COMM_WORLD, &request);
    }

    MPI_Barrier(PETSC_COMM_WORLD);
    printf("Rank = %d finished sending!!!\n", rank);

    if(!localBC.tBC) {
      MPI_Irecv(&_TestV[0][_Index.tbStr+1][0], 1, columntype, NbrRank.t, RecvTag.T2nd, PETSC_COMM_WORLD, &request);
      MPI_Irecv(&_TestV[0][_Index.tbStr][0], 1, columntype,  NbrRank.t, RecvTag.T1st, PETSC_COMM_WORLD, &request);
    }

    if(!localBC.bBC) {
      MPI_Irecv(&_TestV[0][_Index.bbEnd-1][0], 1, columntype, NbrRank.b, RecvTag.B2nd, PETSC_COMM_WORLD, &request);
      MPI_Irecv(&_TestV[0][_Index.bbEnd][0], 1, columntype, NbrRank.b, RecvTag.B1st, PETSC_COMM_WORLD, &request);
    }

    MPI_Wait(&request, &status);


  }


  PetscFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "CleanUpWorks"

PetscErrorCode CleanUpWorks () {
  PetscFunctionBegin;

  MPI_Type_free(&columntype);
  MPI_Type_free(&rowtype);
  MPI_Type_free(&spantype);
  PetscFunctionReturn(0);

}


